#include <YoggieEngine.h>
#include "Application.h"
#include <imgui.h>
#include <backends/imgui_impl_opengl3.h>
#include <backends/imgui_impl_glfw.h>
#include "../../libs/guizmo/ImGuizmo.h"
#include "../../Editor/src/IconsMaterialDesign.h"

namespace YoggieEngine {

	Application::Application(const std::string& name)
		: m_AppName(name)
	{
		appWindow = CreateNativeWindow(1200, 700, m_AppName.c_str());

		// Initialize engine should possibly happen here
		EngineInstrumentation::PerfomanceSamplerInit();

		// Initilize ImGui for this application 
		IMGUI_CHECKVERSION();
		ImGui::CreateContext();
		ImGuiIO& io = ImGui::GetIO();
		io.ConfigFlags |= ImGuiConfigFlags_::ImGuiConfigFlags_ViewportsEnable;
		io.ConfigFlags |= ImGuiConfigFlags_::ImGuiConfigFlags_DockingEnable;
		io.Fonts->AddFontFromFileTTF("build/Debug/Fonts/Roboto-Regular.ttf", 18);
		ImFontConfig config;
		config.MergeMode = true;
		config.GlyphMinAdvanceX = 18.0f;
		static const ImWchar icon_ranges[] = { ICON_MIN_MD , ICON_MAX_MD, 0 };
		io.Fonts->AddFontFromFileTTF("build/Debug/Fonts/MaterialIcons-Regular.ttf", 24, &config, icon_ranges);
		
		ImGui::StyleColorsDark();
		/*
		ImGuiStyle* style = &ImGui::GetStyle();
		ImVec4* colors = style->Colors;

		colors[ImGuiCol_TitleBg] = ImVec4(0.72f, 0.24f, 0.87f, 1.00f);
		colors[ImGuiCol_MenuBarBg] = ImVec4(0.72f, 0.24f, 0.87f, 1.00f);
		colors[ImGuiCol_Tab] = ImVec4(0.53f, 0.09f, 0.67f, 1.00f);

		
		*/
		


		ImGui_ImplGlfw_InitForOpenGL((GLFWwindow*)appWindow->GetHandle(), true);
		ImGui_ImplOpenGL3_Init("#version 450");

		ImGuizmo::SetImGuiContext(ImGui::GetCurrentContext());
		//ImGuizmo::SetOrthographic(true);


		//init_inputSystem(appWindow);
		glfwSetWindowUserPointer((GLFWwindow*)this->appWindow->GetHandle(), this);
		glfwSetKeyCallback((GLFWwindow*)this->appWindow->GetHandle(), HandleKey);

	}



	void Application::HandleKey(GLFWwindow* window, int key, int scancode, int action, int mods) {
		auto app = (Application*)glfwGetWindowUserPointer(window);
		for (auto i = app->AppLayerstack.begin(); i < app->AppLayerstack.end(); i++) {
			if ((*i)->OnKey(key, action)) {
				break;
			}
		}
	}

	void Application::Run() {
		

		for (auto i = AppLayerstack.begin(); i < AppLayerstack.end(); i++) {
			(*i)->OnStartup();
		}
		
		double previous = glfwGetTime();
		double lag = 0.0;

		while (!appWindow->WindowShouldClose()) {
			PollEvents();
			double now = glfwGetTime();
			double elapsed = now - previous;
			previous = now;
			lag += elapsed;


			for (auto i = AppLayerstack.begin(); i < AppLayerstack.end(); i++) {
				(*i)->OnUpdate();
			}

			GuiBegin();
			for (auto i = AppLayerstack.begin(); i < AppLayerstack.end(); i++) {
				(*i)->OnUI();
			}
			GuiEnd();

			SwapBuffers();
		}

		for (auto i = AppLayerstack.begin(); i < AppLayerstack.end(); i++) {
			(*i)->OnDestroy();
		}
	}


	void Application::PollEvents() {
		appWindow->Poll();
	}

	void Application::SwapBuffers()
	{
		appWindow->SwapBuffers();
		glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	}

	void Application::GuiBegin() {
		ImGui_ImplGlfw_NewFrame();
		ImGui_ImplOpenGL3_NewFrame();

		ImGui::NewFrame();
		ImGuizmo::BeginFrame();
		ImGui::DockSpaceOverViewport(ImGui::GetMainViewport());
		
	
	}

	void Application::GuiEnd() {
	
		ImGui::EndFrame();


		ImGui::Render();
		ImGui_ImplOpenGL3_RenderDrawData(ImGui::GetDrawData());


		if (ImGui::GetIO().ConfigFlags & ImGuiConfigFlags_ViewportsEnable)
		{
			GLFWwindow* last_context = glfwGetCurrentContext();
			ImGui::UpdatePlatformWindows();
			ImGui::RenderPlatformWindowsDefault();
			glfwMakeContextCurrent(last_context);
		}
	}

	void Application::PushLayer(Layer* layer)
	{
		AppLayerstack.PushLayer(layer);
		layer->OnAttach();
	}

	Application::~Application() {

		ImGui_ImplOpenGL3_Shutdown();
		ImGui_ImplGlfw_Shutdown();
		ImGui::DestroyContext();

		delete appWindow;

	}
}
