#include <YoggieEngine.h>

#include "Renderer.h"
#include "../Scene/Components.h"
#include "../Graphics/Memory/Buffer.h"
#include "../Graphics/Memory/VertexArray.h"
#include "Renderable.h"
#include "Memory/Framebuffer.h"
#include "../Scene/Components.h"
#include"../Scene/Scene.h"
#include "Primitives/Camera.h"
#include "OpenglAPI.h"

namespace YoggieEngine {
unsigned int quadVAO = 0;
unsigned int quadVBO = 0;
// vegetation test
std::vector<glm::vec3> vegetation = {
	glm::vec3(-1.5f, 0.0f, -0.48f),
	glm::vec3(1.5f, 0.0f, 0.51f),
	glm::vec3(0.0f, 0.0f, 0.7f),
	glm::vec3(-0.3f, 0.0f, -2.3f)
};

unsigned int transparentVAO, transparentVBO;
Texture grassTexture;

float skyboxVertices[]{
	// positions          
	-1.0f,  1.0f, -1.0f,
	-1.0f, -1.0f, -1.0f,
	 1.0f, -1.0f, -1.0f,
	 1.0f, -1.0f, -1.0f,
	 1.0f,  1.0f, -1.0f,
	-1.0f,  1.0f, -1.0f,

	-1.0f, -1.0f,  1.0f,
	-1.0f, -1.0f, -1.0f,
	-1.0f,  1.0f, -1.0f,
	-1.0f,  1.0f, -1.0f,
	-1.0f,  1.0f,  1.0f,
	-1.0f, -1.0f,  1.0f,

	 1.0f, -1.0f, -1.0f,
	 1.0f, -1.0f,  1.0f,
	 1.0f,  1.0f,  1.0f,
	 1.0f,  1.0f,  1.0f,
	 1.0f,  1.0f, -1.0f,
	 1.0f, -1.0f, -1.0f,

	-1.0f, -1.0f,  1.0f,
	-1.0f,  1.0f,  1.0f,
	 1.0f,  1.0f,  1.0f,
	 1.0f,  1.0f,  1.0f,
	 1.0f, -1.0f,  1.0f,
	-1.0f, -1.0f,  1.0f,

	-1.0f,  1.0f, -1.0f,
	 1.0f,  1.0f, -1.0f,
	 1.0f,  1.0f,  1.0f,
	 1.0f,  1.0f,  1.0f,
	-1.0f,  1.0f,  1.0f,
	-1.0f,  1.0f, -1.0f,

	-1.0f, -1.0f, -1.0f,
	-1.0f, -1.0f,  1.0f,
	 1.0f, -1.0f, -1.0f,
	 1.0f, -1.0f, -1.0f,
	-1.0f, -1.0f,  1.0f,
	 1.0f, -1.0f,  1.0f
};


Renderer::Renderer() :
	m_framebuffer(800, 600),
	gBufferShader("build/Debug/Shaders/deferred/geometry.vert", "build/Debug/Shaders/deferred/geometry.frag"),
	lightingPassShader("build/Debug/Shaders/deferred/lightPass.vert", "build/Debug/Shaders/deferred/lightPass.frag"),
	SkyboxShader("build/Debug/Shaders/Cubemaps/Skybox.vert", "build/Debug/Shaders/Cubemaps/Skybox.frag"),
	BlendingShader("build/Debug/Shaders/forward/Blending.vert", "build/Debug/Shaders/forward/Blending.frag"),
	forwardShader("build/Debug/Shaders/forward/geometry.vert", "build/Debug/Shaders/forward/geometry.frag"),
	postProcessingShader("build/Debug/Shaders/forward/postprocessing.vert", "build/Debug/Shaders/forward/postprocessing.frag")
{
	width = 800;
	height = 600;
	glEnable(GL_DEPTH_TEST);
	CreateGBuffer();

	std::vector<std::string> faces{
			"build/Debug/skybox/Open_Water/right.jpg",
			"build/Debug/skybox/Open_Water/left.jpg",
			"build/Debug/skybox/Open_Water/top.jpg",
			"build/Debug/skybox/Open_Water/bottom.jpg",
			"build/Debug/skybox/Open_Water/front.jpg",
			"build/Debug/skybox/Open_Water/back.jpg"
	};

	skybox = CubeMap(faces);

}


void Renderer::CreateGBuffer() {
	// Deferred Rendering 
	glGenFramebuffers(1, &gBuffer);
	glBindFramebuffer(GL_FRAMEBUFFER, gBuffer);

	// - Position Color Buffer ;
	glGenTextures(1, &gPosition);
	glBindTexture(GL_TEXTURE_2D, gPosition);
	glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA16F, width, height, 0, GL_RGBA, GL_FLOAT, NULL);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
	glFramebufferTexture2D(GL_FRAMEBUFFER, GL_COLOR_ATTACHMENT0, GL_TEXTURE_2D, gPosition, 0);

	// - normal color buffer ;
	glGenTextures(1, &gNormal);
	glBindTexture(GL_TEXTURE_2D, gNormal);
	glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA16F, width, height, 0, GL_RGBA, GL_FLOAT, NULL);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
	glFramebufferTexture2D(GL_FRAMEBUFFER, GL_COLOR_ATTACHMENT1, GL_TEXTURE_2D, gNormal, 0);

	// - Color + Specular buffer ;
	glGenTextures(1, &gColorSpec);
	glBindTexture(GL_TEXTURE_2D, gColorSpec);
	glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, width, height, 0, GL_RGBA, GL_UNSIGNED_BYTE, NULL);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
	glFramebufferTexture2D(GL_FRAMEBUFFER, GL_COLOR_ATTACHMENT2, GL_TEXTURE_2D, gColorSpec, 0);


	unsigned int attachments[3] = { GL_COLOR_ATTACHMENT0, GL_COLOR_ATTACHMENT1 , GL_COLOR_ATTACHMENT2 };
	glDrawBuffers(3, attachments);

	// Create and attach a depth buffer 
	glGenRenderbuffers(1, &gDepth);
	glBindRenderbuffer(GL_RENDERBUFFER, gDepth);
	glRenderbufferStorage(GL_RENDERBUFFER, GL_DEPTH_COMPONENT, width, height);
	glFramebufferRenderbuffer(GL_FRAMEBUFFER, GL_DEPTH_ATTACHMENT, GL_RENDERBUFFER, gDepth);

	if (glCheckFramebufferStatus(GL_FRAMEBUFFER) != GL_FRAMEBUFFER_COMPLETE)
	{
		spdlog::critical("Framebuffer not complete {}{}", __FILE__, __LINE__);
	}

	glBindFramebuffer(GL_FRAMEBUFFER, 0);

}


void Renderer::Render(Scene& scene , Camera MainCamera){
	int oldviewport[4];
	glGetIntegerv(GL_VIEWPORT, oldviewport);
	glViewport(0, 0, width, height);


	glBindFramebuffer(GL_FRAMEBUFFER, gBuffer);
	glClearColor(0.0f, 0.0f, 0.0f, 1.0f);
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	// Render skybox if the scene has one 

	SkyboxShader.Use();
	SkyboxShader.setUniformMat4("projection", MainCamera.getProjection(width, height));
	SkyboxShader.setUniformMat4("view", glm::inverse( glm::mat4(glm::mat3(MainCamera.getTransform()))));

	if (!skyboxVAO) {
		unsigned int VBO;
		glGenVertexArrays(1, &skyboxVAO);
		glGenBuffers(1, &VBO);
		glBindVertexArray(skyboxVAO);
		glBindBuffer(GL_ARRAY_BUFFER, VBO);
		glBufferData(GL_ARRAY_BUFFER, sizeof(skyboxVertices), &skyboxVertices, GL_STATIC_DRAW);
		glEnableVertexAttribArray(0);
		glVertexAttribPointer(0, 3, GL_FLOAT, GL_FALSE, 3 * sizeof(float), (void*)0);

	}

	OpenGLApi::DrawCubeMap(skyboxVAO, skybox);




	auto renderables = scene.getReg().view<TransformComponent, Render3DComponent>();
	gBufferShader.Use();
	for (auto renderable : renderables) {
		auto entity = Entity(renderable, &scene);
		auto& renderComponent = entity.GetComponent<Render3DComponent>();
		auto& transform = entity.GetComponent<TransformComponent>();

		// Geometry pass 
		gBufferShader.setUniformVec3("Color", renderComponent.color);
		gBufferShader.setUniformMat4("Model", transform.GetTransform());
		gBufferShader.setUniformMat4("View", glm::inverse(MainCamera.getTransform()));
		gBufferShader.setUniformMat4("Projection", MainCamera.getProjection(width, height));
		OpenGLApi::DrawTriangles(renderComponent);


	}

	// Light pass
	glBindFramebuffer(GL_FRAMEBUFFER, m_framebuffer.GetId());
	lightingPassShader.Use();
	lightingPassShader.setUniformInt("gPosition", 0);
	lightingPassShader.setUniformInt("gNormal", 1);
	lightingPassShader.setUniformInt("gColorSpec", 2);

	glActiveTexture(GL_TEXTURE0);
	glBindTexture(GL_TEXTURE_2D, gPosition);

	glActiveTexture(GL_TEXTURE1);
	glBindTexture(GL_TEXTURE_2D, gNormal);

	glActiveTexture(GL_TEXTURE2);
	glBindTexture(GL_TEXTURE_2D, gColorSpec);

	auto lights = scene.getReg().view<LightComponent, TransformComponent>();
	unsigned int lightnr = 0;
	for (auto light : lights) {
		auto lightComponent = Entity(light, &scene).GetComponent<LightComponent>();
		auto transformComponent = Entity(light, &scene).GetComponent<TransformComponent>();

		auto name = "lights[" + std::to_string(lightnr) + "]";
		lightingPassShader.setUniformVec3(name + ".Position", transformComponent.Position);
		lightingPassShader.setUniformVec3(name + ".Color", lightComponent.Color);

		const float linear = 0.7f;
		const float quadratic = 1.8f;
		lightingPassShader.setUniformFloat(name + ".Linear", linear);
		lightingPassShader.setUniformFloat(name + ".Quadratic", quadratic);

		lightnr++;

	}

	if (quadVAO == 0)
	{
		float quadVertices[] = {
			-1.0f,	 1.0f, 0.0f, 0.0f, -1.0f,
			-1.0f,	-1.0f, 0.0f, 0.0f, 0.0f,
			 1.0f,	 1.0f, 0.0f, 1.0f, -1.0f,
			 1.0f,	-1.0f, 0.0f, 1.0f, 0.0f,
		};
		unsigned int quadVBO;
		// setup plane VAO ;
		glGenVertexArrays(1, &quadVAO);
		glGenBuffers(1, &quadVBO);
		glBindVertexArray(quadVAO);
		glBindBuffer(GL_ARRAY_BUFFER, quadVBO);
		glBufferData(GL_ARRAY_BUFFER, sizeof(quadVertices), &quadVertices, GL_STATIC_DRAW);
		glEnableVertexAttribArray(0);
		glVertexAttribPointer(0, 3, GL_FLOAT, GL_FALSE, 5 * sizeof(float), (void*)0);
		glEnableVertexAttribArray(1);
		glVertexAttribPointer(1, 2, GL_FLOAT, GL_FALSE, 5 * sizeof(float), (void*)(3 * sizeof(float)));
	}

	glBindVertexArray(quadVAO);
	glDrawArrays(GL_TRIANGLE_STRIP, 0, 4);
	glBindVertexArray(0);


	// Copy GBuffer
	glBindFramebuffer(GL_READ_FRAMEBUFFER, gBuffer);
	glBindFramebuffer(GL_DRAW_BUFFER, m_framebuffer.GetId());
	glBlitFramebuffer(0, 0, width, height, 0, 0, width, height, GL_DEPTH_BUFFER_BIT, GL_NEAREST);


	if (transparentVAO == 0) {
		unsigned int transparentVBO;
		float transparentVertices[] = {
			// positions         // texture Coords (swapped y coordinates because texture is flipped upside down)
			0.0f,  0.5f,  0.0f,  0.0f,  1.0f,
			0.0f, -0.5f,  0.0f,  0.0f,  0.0f,
			1.0f, -0.5f,  0.0f,  1.0f,  0.0f,

			0.0f,  0.5f,  0.0f,  0.0f,  1.0f,
			1.0f, -0.5f,  0.0f,  1.0f,  0.0f,
			1.0f,  0.5f,  0.0f,  1.0f,  1.0f
		};

		glGenVertexArrays(1, &transparentVAO);
		glGenBuffers(1, &transparentVBO);
		glBindVertexArray(transparentVAO);
		glBindBuffer(GL_ARRAY_BUFFER, transparentVBO);
		glBufferData(GL_ARRAY_BUFFER, sizeof(transparentVertices), transparentVertices, GL_STATIC_DRAW);
		glEnableVertexAttribArray(0);
		glVertexAttribPointer(0, 3, GL_FLOAT, GL_FALSE, 5 * sizeof(float), (void*)0);
		glEnableVertexAttribArray(1);
		glVertexAttribPointer(1, 2, GL_FLOAT, GL_FALSE, 5 * sizeof(float), (void*)(3 * sizeof(float)));
		glBindVertexArray(0);

	}

	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

	BlendingShader.Use();
	glBindVertexArray(transparentVAO);
	glActiveTexture(GL_TEXTURE0);
	glBindTexture(GL_TEXTURE_2D, grassTexture.GetID());

	BlendingShader.setUniformMat4("V", glm::inverse(MainCamera.getTransform()));
	BlendingShader.setUniformMat4("P", MainCamera.getProjection(width, height));
	for (unsigned int i = 0; i < vegetation.size(); i++) {

		auto translation = glm::translate(glm::mat4(1.0f), vegetation[i]);
		BlendingShader.setUniformMat4("M", translation);
		glDrawArrays(GL_TRIANGLES, 0, 6);

	}

	glBindVertexArray(0);

	glDisable(GL_BLEND);





	glBindFramebuffer(GL_FRAMEBUFFER, 0);
	glViewport(oldviewport[0], oldviewport[1], oldviewport[2], oldviewport[3]);
}


Renderer::~Renderer(){}





void Renderer::setCurrentFrameBuffer(const Framebuffer& fb)
{
	m_framebuffer = fb;
}

void Renderer::setClearColor(const glm::vec3& ClearColor)
{
	m_clearColor = ClearColor;
}

}