#pragma once
#include "GLFW/glfw3.h"
#include <vector>
#include "../PerfCounter.h"
#include "../Scene/Scene.h"
#include "Memory/Framebuffer.h"

namespace YoggieEngine {

	struct RendererConfig {
		int ScreenWidth, ScreenHeight;
		glm::vec3 ClearColor;
		bool DepthTest; 
	};

	class Renderer {
	public:
		Renderer();
		~Renderer();


		// Forward pass
		/*
		forwardShader.Use();
		forwardShader.setUniformVec3("Color", renderComponent.color);
		forwardShader.setUniformMat4("M", transform.LocalTransform);
		forwardShader.setUniformMat4("V", MainCamera.view);
		forwardShader.setUniformMat4("P", MainCamera.projection);

		OpenGLApi::DrawTriangles(renderComponent);
		*/

		void Render(Scene& scene, Camera MainCamera);
			
		void setCurrentFrameBuffer(const Framebuffer& fb);
		void setClearColor(const glm::vec3& ClearColor);


		Framebuffer& getCurrentFrameBuffer() { return m_framebuffer; }

	private:
		Framebuffer m_framebuffer;
		int width, height;
		glm::vec3 m_clearColor;
		bool m_depthTest;

		void CreateGBuffer();

		Shader forwardShader;

		// deferred rending parameters
		unsigned int gBuffer, gPosition, gNormal, gColorSpec, gDepth;
		Shader lightingPassShader;
		Shader gBufferShader;
		Shader SkyboxShader;
		Shader postProcessingShader;

		// blending
		Shader BlendingShader;
		CubeMap skybox;
		Texture grassTexture = Texture("build/Debug/Texture/grass.png");

		unsigned int transparentVAO = 0;
		unsigned int skyboxVAO = 0;
		unsigned int quadVAO = 0;
		float skyboxVertices[36*3]{
			// positions          
			-1.0f,  1.0f, -1.0f,
			-1.0f, -1.0f, -1.0f,
			 1.0f, -1.0f, -1.0f,
			 1.0f, -1.0f, -1.0f,
			 1.0f,  1.0f, -1.0f,
			-1.0f,  1.0f, -1.0f,
			-1.0f, -1.0f,  1.0f,
			-1.0f, -1.0f, -1.0f,
			-1.0f,  1.0f, -1.0f,
			-1.0f,  1.0f, -1.0f,
			-1.0f,  1.0f,  1.0f,
			-1.0f, -1.0f,  1.0f,
			 1.0f, -1.0f, -1.0f,
			 1.0f, -1.0f,  1.0f,
			 1.0f,  1.0f,  1.0f,
			 1.0f,  1.0f,  1.0f,
			 1.0f,  1.0f, -1.0f,
			 1.0f, -1.0f, -1.0f,
			-1.0f, -1.0f,  1.0f,
			-1.0f,  1.0f,  1.0f,
			 1.0f,  1.0f,  1.0f,
			 1.0f,  1.0f,  1.0f,
			 1.0f, -1.0f,  1.0f,
			-1.0f, -1.0f,  1.0f,
			-1.0f,  1.0f, -1.0f,
			 1.0f,  1.0f, -1.0f,
			 1.0f,  1.0f,  1.0f,
			 1.0f,  1.0f,  1.0f,
			-1.0f,  1.0f,  1.0f,
			-1.0f,  1.0f, -1.0f,
			-1.0f, -1.0f, -1.0f,
			-1.0f, -1.0f,  1.0f,
			 1.0f, -1.0f, -1.0f,
			 1.0f, -1.0f, -1.0f,
			-1.0f, -1.0f,  1.0f,
			 1.0f, -1.0f,  1.0f
		};
		std::vector<glm::vec3> vegetation = {
	glm::vec3(-1.5f, 0.0f, -0.48f),
	glm::vec3(1.5f, 0.0f, 0.51f),
	glm::vec3(0.0f, 0.0f, 0.7f),
	glm::vec3(-0.3f, 0.0f, -2.3f)
		};

	};
}
