#pragma once
#include "Entity.h"

#include <glm/glm.hpp>
#include <glm/gtx/quaternion.hpp>

namespace YoggieEngine {
	struct IdentifierComponent {
		std::string name;
	};


	struct TransformComponent {
		glm::vec3 Position = glm::vec3(0.0f, 0.0f, 0.0f);
		glm::vec3 Rotation = glm::vec3(0.0f , 0.0f, 0.0f);
		glm::vec3 Scale = glm::vec3(1.0f);
		
		glm::mat4 LocalTransform = glm::mat4(1.0f);

		glm::mat4 GetTransform() const {
			glm::mat4 rotation = glm::toMat4(glm::quat(Rotation));

			auto transform = glm::translate(glm::mat4(1.0f), Position)* rotation 
				* glm::scale(glm::mat4(1.0f), Scale);

			

			return transform;
		}

		void Decompose(glm::mat4& transformationMatrix) {

			

			auto& a = transformationMatrix[0][0];
			auto& b = transformationMatrix[1][0];
			auto& c = transformationMatrix[2][0];
			auto& d = transformationMatrix[3][0];

			auto& e = transformationMatrix[0][1];
			auto& f = transformationMatrix[1][1];
			auto& g = transformationMatrix[2][1];
			auto& h = transformationMatrix[3][1];

			auto& i = transformationMatrix[0][2];
			auto& j = transformationMatrix[1][2];
			auto& k = transformationMatrix[2][2];
			auto& l = transformationMatrix[3][2];

			Position = glm::vec3(transformationMatrix[3]);
			
			// Remove the position from the matrix
			d = 0;
			h = 0;
			l = 0;
		
			auto sx = glm::sqrt(a*a + e*e + i*i);
			auto sy = glm::sqrt(b*b + f*f + j*j);
			auto sz = glm::sqrt(c*c + g*g + k*k);

			Scale = glm::vec3(sx, sy, sz);
	
			// Remove the scale from the matrix;
			a/= sx;
			e /= sx;
			i /= sx;

			b /= sy;
			f /= sy;
			j /= sy;

			c /= sz;
			g /= sz;
			k /= sz;
			

			auto w = glm::sqrt(1 + transformationMatrix[0][0] + transformationMatrix[1][1] + transformationMatrix[2][2]) / 2;


			auto x = (transformationMatrix[2][1] - transformationMatrix[1][2]) / (4 * w);
			auto y = (transformationMatrix[0][2] - transformationMatrix[2][0]) / (4 * w);
			auto z = (transformationMatrix[1][0] - transformationMatrix[0][1]) / (4 * w);


			auto rot = glm::quat(w, x, y, z);
			Rotation = glm::eulerAngles(rot);


		}


	};

	struct RelationComponent {
		Entity Parent;

		RelationComponent(Entity& entity) 
		{
			Parent = entity;
		}

	};


	struct LightComponent {
		glm::vec3 Color = glm::vec3(1.0f, .5f, .5f);
	};


	struct CameraComponent {
		glm::vec3 Position = glm::vec3(0);
		glm::vec3 Rotation = glm::vec3(0);
		float Zoom ;
	};


	struct ScriptComponent {
		std::string file; // TODO : replace with proper properties
	};


	struct RigidBody {
		float Mass;
		// Physics Material
	};

	struct BoxCollider {
		glm::vec3 extents;
	};

	struct PlaneCollider {

	};


	struct Render3DComponent {

		unsigned int VAO = 0;
		unsigned int IBO = 0;
		Mesh mesh;
		// TODO: becomes a material 
		glm::vec3 color;
		bool isStatic;

		Render3DComponent()
			: color(glm::vec3(1.0f, 0.0f, 0.0f)) , isStatic(true)
		{ 
		}
	};

	struct Render2DComponent {
		glm::vec3 Colour;
		
	};
};